#  common_functions_MA.R
#  common functions related to MA plots and Rotation/Lowess Normalization
#----------------------------- loess normalization -------
# new: weights =0 for the outliers f.out*sd(M) away from median(M)
#-----------------------------------------------------------------
library(spatstat)
library(MASS)
#MA <- ppp(A,M)

plotMA <- function(M,A, sample=3000,pch=".",main="ChIP-chip",
                   xlim=NULL,ylim=NULL,xlab="A (intensity)",ylab="M (log ratio)",
                   index=list(index.X.F,index.2L.F), legend=NULL,
                   cex=1.5, cex.lab=1.5, cex.axis=1.5,
                   plot.contour=FALSE,levels=seq(0.01,0.05,0.01),n=30){
  if (is.null(index)){
    index.1 <- 1:length(A)
    index.2 <- NULL
  }else{
    index.1 <- index[[1]]
    index.2 <- index[[2]]
  }
  if (is.null(xlim))   xlim <- c(floor(min(A)),ceiling(max(A)))
  if (is.null(ylim))   ylim <- c(floor(min(M)),ceiling(max(M)))
  if (is.null(sample)){
    id.1 <- index.1
    id.2 <- index.2
  }else{
    if (length(index.1)> sample) {
      id.1 <- sample(index.1,size=sample)
    }else{
      id.1 <- index.1
    }
    if (length(index.2)> sample){
      id.2 <- sample(index.2,size=sample)
    }else{
      id.2 <- index.2
    }
  }
  plot(A[id.1],M[id.1],pch=pch,col=2, xlab=xlab,ylab=ylab,xlim=xlim,ylim=ylim,
       main=main,cex.lab=cex.lab,cex.axis=cex.axis)
  if (length(id.2)>0) {
    points(A[id.2],M[id.2],pch=pch,col=4)
    if (is.null(legend)) {
      legend <- c("Chr.X", "Chr.2L")
      legend("bottomleft",legend=legend,text.col=c(2,4))
    }
  }
  
  if (plot.contour == TRUE){
#    levels <- seq(0.01,0.10,0.01)
    f <- get.contour(x=A,y=M,n=n)
    contour(f,nlevels=20,levels=levels,col=1,add=TRUE, cex=cex,cex.lab=cex.lab)
  }
}

set.weightsMA <- function(MA,f.out=2,minsize=1){
  if (class(MA) =="ppp"){
    M <- MA$y; m.M <- median(M); mad.M <- mad(M)
    A <- MA$x; m.A <- median(A); mad.A <- mad(A)
    print(c("mad.M", mad.M))
    print(c("mad.A", mad.A))
    mad.ave <- (mad.M+mad.A)/2
    M0 <- (M-m.M)/mad.ave
    A0 <- (A-m.A)/mad.ave
    d2 <- M0^2+A0^2
  }else{
    M <- MA; m.M  <- median(M); mad.M <- mad(M)
    M0 <- (M-m.M)/mad.M
    d2 <- M0^2
  }
  
  f.out2 <- f.out^2
  w <- as.numeric(d2<=f.out2)
  length(w)
  return(w)
}
set.weightsMA.old <- function(MA,f.out=2,minsize=1){
  if (class(MA) =="ppp"){
    M <- MA$y; m.M <- median(M); mad.M <- mad(M)
    A <- MA$x; m.A <- median(A); mad.A <- mad(A)
    print(c("mad.M", mad.M))
    print(c("mad.A", mad.A))    
    M0 <- (M-m.M)/mad.M
    A0 <- (A-m.A)/mad.A
    d2 <- M0^2+A0^2
  }else{
    M <- MA; m.M  <- median(M); mad.M <- mad(M)
    M0 <- (M-m.M)/mad.M
    d2 <- M0^2
  }
  
  f.out2 <- f.out^2
  w <- as.numeric(d2<=f.out2)
  length(w)
  return(w)
}

set.weights <- function(M,f.out=2,minsize=5){
#  print("set cluter outliers' weights  to zero")
  n <- length(M)
  w <- rep(1,n)
  m <- median(M)
  delta <- sd(M)

  cObj0 <- get.cutoff.clusters( M,  m+f.out*delta, minsize)
  cObj1 <- get.cutoff.clusters(-M, -m+f.out*delta, minsize)
 
  if(length(cObj0$size)!=0){
    for (j in 1:length(cObj0$size)){
      ind.temp <- cObj0$begin[j]:cObj0$end[j]
      w[ind.temp] <-0
    }
  }

  if(length(cObj1$size)!=0){
    for (j in 1:length(cObj1$size)){
      ind.temp <- cObj1$begin[j]:cObj1$end[j]
      w[ind.temp] <-0
    }
  }
#  print(length(which(w==0))/length(w))
  return(w)
}

get.lm <- function(MA,f.out=2){
  lm.MA <- lm(MA$y~MA$x)
  c0 <- coef(lm.MA)
  M <- MA$y; A <- MA$x
  weights1 <- set.weights(M,f.out=f.out)
  ind1 <- which(weights1==0)
  if (length(ind1)>0){
    M1 <- M[-ind1]
    A1 <- A[-ind1]
  }else{
    M1 <- M
    A1 <- A
  }
  lm.MA1 <- lm(M1~A1)
  c1 <- coef(lm.MA1)
#  slope <- c1[2]
#  return(slope)
#return(list(a=c1[1],b=c1[2]))
  return(c1)
}

rotateMA <- function(MA, f.out=2, iteration=3,lag=0,angle=NULL){
  if (is.null(angle)){
    angles <- numeric(iteration)
  }else{
    iteration=1
  }
  MA0 <- MA
  for (j in 1:iteration){    
#    slope <- get.lm(MA0,f.out=f.out)[2]
#    angle[j] <- -atan(slope)
    if (is.null(angle)){
      angle0 <- angles[j] <- get.angle(MA0,f.out=f.out,lag=lag)
    }else{
      angle0 <- angle
    }
    o <- c(median(MA0$x),median(MA0$y))  # center of rotation
    MA.s <- shift(MA0,-o)                # shift to center of rotation
    MA.r <- rotate(MA.s,angle0)        # rotate
    MA0 <- shift(MA.r,o)                 # shift back
  }
  if (is.null(angle)){
    sum.angles <- sum(angles)
  }else{
    sum.angles <- angles <- angle
  }

  # shift median of log ratio to zero
  #MA0$y <- MA0$y-median(MA0$y)
  
  #shift the maximum density of log ratio to zero
  MA0$y <- MA0$y-density.max(MA0$y)
  
  return(list(MA=MA0, angle=sum.angles,angles=angles))
}

get.MA.rl <- function(MA,f.out=2,iteration=3,lag=0,angle=NULL){
  
#  MA$y <- MA$y-median(MA$y)
  MA.r <- rotateMA(MA,f.out=f.out,iteration=iteration,lag=lag,angle=angle); rMA <- MA.r$MA

  
  weights <- set.weights(rMA$y,f.out)
  fit <- loess(rMA$y ~ rMA$x, span = 0.3, degree = 1,
               surface = "interpolate", family = "symmetric",
               trace.hat = "approximate", iterations = 4,
               weights=weights)
  
  MA.lowess <- ppp(rMA$x,rMA$y-fit$fitted)

  #------- shift the max density of log ratio to zero -- 
  MA.lowess$y <- MA.lowess$y-density.max(MA.lowess$y)
  
  return(list(MA = MA, MA.r = MA.r, fit=fit, MA.l=MA.lowess))
}

plot.MA.rl <- function(MA.rl, main="", f.out=2,cex=0.8,sample =3000,p.list=NULL,index=list(index.X.F,index.2L.F)){
  print(sample)
  MA   <- MA.rl$MA
  MA.r <- MA.rl$MA.r
  rMA  <- MA.r$MA
  fit  <- MA.rl$fit
  MA.lowess <- MA.rl$MA.l
  if (is.null(index)){
    index.X.F <- 1:length(MA$x)
    index.2L.F <- NULL
  }else{
    index.X.F <- index[[1]]
    index.2L.F <- index[[2]]
  }
  
  if (is.null(sample)){
    print(sample)
    ind <- union(index.X.F,index.2L.F)
  }else{
    ind <- union(sample(index.X.F,size=sample),sample(index.2L.F,size=sample))
  }
  length(ind)
  p.list0 <- NULL
  if (!is.null(p.list)){
    for (j in 1:length(p.list)){
      p.list0[[j]] <- intersect(p.list[[j]],ind)
    }
  }

  plotMA(sample=sample,M=MA$y, A=MA$x,main=main,index=index )
  
  if (!is.null(p.list0)){
    for (j in 1:length(p.list0)){
      points(x=MA$x[p.list0[[j]]],y=MA$y[p.list0[[j]]],cex=0.5,col=2*j-1)
    }
  }
#  legend("bottomcent",legend="d")

  d.M <- diff(MA$y,lag=lag); d.A <- diff(MA$x,lag=lag);lim <- range(c(d.M,d.A))
  plotMA(sample=sample,M=d.M, A=d.A,xlim=lim,ylim=lim,
         xlab="lagged diff. of A", ylab ="lagged diff. of M",
         main=paste("lag=",lag,sep=""),index=index)
  b <- -tan(MA.r$angle)
  abline(a=0,b=b,lty=2,col="green")
  
  plotMA(sample=sample,M=rMA$y, A=rMA$x, main="Rotation",index=index)
  if (!is.null(p.list0)){
    for (j in 1:length(p.list0)){
      points(x=rMA$x[p.list0[[j]]],y=rMA$y[p.list0[[j]]],cex=0.5,col=2*j-1)
    }
  }

  abline(h=0,lty=2,col="green")
  points(rMA$x[ind],fit$fitted[ind],pch=".",col="grey")
  angles <- paste(as.numeric(round(MA.r$angles*4/pi,2)), collapse="; ")
  legend <- paste("Angle: ", round(MA.r$angle*4/pi,2), " [",angles,"]","*pi/4", sep="")
#  legend("topleft",legend=legend,text.col=1,cex=cex)
  
  plotMA(sample=sample,A=MA.lowess$x,M=MA.lowess$y,main="Loess normalization",index=index)
  
  if (!is.null(p.list0)){
    for (j in 1:length(p.list0)){
      points(x=MA.lowess$x[p.list0[[j]]],y=MA.lowess$y[p.list0[[j]]],cex=0.5,col=2*j-1)
    }
  }
  
  legend=paste("Zero weight for outliers (", f.out,"*sd)", sep="");legend("topleft",legend=legend,text.col=1,cex=cex)
  abline(h=0,col="grey")
}

plotACF <- function(x,pch=".",cex=1,main="Auto Correlation Function", log=FALSE,
                   xlim=NULL,ylim=NULL,xlab=NULL,ylab=NULL,lag.max=NULL,resolution=100){
  acf.all <- acf(x,plot=FALSE)
  acf.X <- acf(x[index.X.F],plot=FALSE)
  acf.2L<- acf(x[index.2L.F],plot=FALSE)  
  if (is.null(lag.max))
    lag.max <- min(max(acf.all$lag),max(acf.X$lag),max(acf.2L$lag))
  ind.lag <- 1+0:lag.max
  acf.m <- cbind(acf.X$acf[ind.lag],acf.2L$acf[ind.lag],acf.all$acf[ind.lag])

  if (is.null(xlab))
    xlab <- "lag"
  if (is.null(ylab))
    ylab <- "Auto Correlation"
  if (is.null(resolution))
    resolution =1
  ind.lag <- ind.lag*resolution
  
  if (log == TRUE) {
    acf.m <- log(acf.m)
    ylab <- "Log(acf)"
  }
  if (is.null(xlim))
    xlim <- c(0,lag.max)
  if (is.null(ylim))
    ylim <- range(acf.m)

  acf.m
  col <- c("red","blue","black")
  plot(ind.lag, acf.m[,1], ylim=ylim, xlab=xlab, ylab=ylab,main=main,col=col[1])
  points(ind.lag, acf.m[,2],col=col[2])
  points(ind.lag, acf.m[,3],col=col[3])
  legend("topright",legend=c("X chromosome","2L chromosome","Both chromosomes"),text.col=col)
  #abline(v=7*resolution,lty=2,col="green");
  abline(v=8*resolution,lty=2,col="green")
}

get.angle <- function(MA, f.out=2,lag=0){
  if (lag==0){
    M <- MA$y; A <- MA$x
  }else{
    M <- diff(MA$y,lag=lag); A <- diff(MA$x,lag=lag)
  }

#  weights1 <- set.weights(M,f.out=f.out)
#  weights2 <- set.weights(M,f.out=f.out)
#  ind1 <- unique(union(which(weights1==0),which(weights2==0)))
  weights1 <- set.weightsMA(ppp(A,M),f.out=f.out)
  ind1 <- which(weights1==0)
  if (length(ind1)>0){
    M1 <- M[-ind1]
    A1 <- A[-ind1]
  }else{
    M1 <- M
    A1 <- A
  } 
  c <- cov(cbind(A1,M1))
  
  svd.c <- svd(c)
#  svd.c
  angle <- asin(svd.c$u[1,2])
  angle*4/pi
  return(angle)
}

#f.out=2;iteration=3;lag=8
get.MA.rl.all <- function(M,A, f.out=2, iteration=3, lag=8, plot=FALSE,angles=NULL){  
  if (is.matrix(A)){
    A.matrix <- A
  }else{
    A.matrix <- matrix(A,ncol=1)
  }
  
  if (is.matrix(M)){
    M.matrix <- M
  }else{
    M.matrix <- matrix(M,ncol=1)
  }
  
  print(  n <- ncol(A.matrix))
  angles0 <- numeric(n)
  if (n != ncol(M.matrix)){
    print("M and A doesn't match in length!!")
    return()
  }else{
    M.rl <- matrix(0, nrow=nrow(M),ncol=n); colnames(M.rl) <- colnames(M)
    A.r <- M.r <- A.rl <- M.rl
 
    if (plot==TRUE)
      par(mfrow=c(n,4))
    
    for (j in 1:ncol(A.matrix)){
      print(j)
      dim(M.matrix)
      MA <- ppp(A.matrix[,j],M.matrix[,j])
      MA.rl <- get.MA.rl(MA,f.out,iteration,lag,angles[j])
      
      M.r[,j] <- MA.rl$MA.r$MA$y
      A.r[,j] <- MA.rl$MA.r$MA$x
      
      angles0[j] <-       MA.rl$MA.r$angle

      M.rl[,j] <- MA.rl$MA.l$y
      A.rl[,j] <- MA.rl$MA.l$x
      
      if (plot==TRUE){
        main=colnames(M.rl)[j];  plot.MA.rl(MA.rl,main=main,f.out,cex=0.7)   
      }
    }
  }
  return(list(M.r =M.r, A.r=A.r, M.rl=M.rl,A.rl=A.rl,angles=angles0))
}

get.contour <- function(x,y,n=30){
  xlim <- range(x); ylim <- range(y);
  #lims <- c(xlim,ylim)
  lim <- range(c(x,y))
  lims <- rep(lim,2)
  f	<- kde2d(x,y,n=n,lims=lims,
		h=rep(0.6,2))
  return(f)
}
